import sqlparse
from django.db import DatabaseError, Error, DataError, OperationalError, \
    IntegrityError, InternalError, ProgrammingError, NotSupportedError, InterfaceError
from django.db.backends.sqlite3.base import DatabaseWrapper as SQLiteDatabaseWrapper
from django.db.backends.sqlite3.client import DatabaseClient as SQLiteDatabaseClient
from django.db.backends.sqlite3.creation import DatabaseCreation as SQLiteDatabaseCreation
from django.db.backends.sqlite3.features import DatabaseFeatures as SQLiteDatabaseFeatures
from django.db.backends.sqlite3.introspection import DatabaseIntrospection as SQLiteDatabaseIntrospection
from django.db.backends.sqlite3.operations import DatabaseOperations as SQLiteDatabaseOperations
from django.db.backends.sqlite3.schema import DatabaseSchemaEditor as SQLiteDatabaseSchemaEditor
from django.db.models.functions import TruncDate, TruncTime, TruncYear, TruncQuarter, TruncMonth, TruncWeek, TruncDay, TruncHour, TruncMinute, TruncSecond
from django.db.models.sql.compiler import SQLCompiler


class CFDatabaseIntrospection(SQLiteDatabaseIntrospection):
    pass


class CFDatabaseCreation(SQLiteDatabaseCreation):
    pass


class CFDatabaseClient(SQLiteDatabaseClient):
    pass


class CFDatabaseSchemaEditor(SQLiteDatabaseSchemaEditor):
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            for sql in self.deferred_sql:
                self.execute(sql)
        if self.atomic_migration:
            self.atomic.__exit__(exc_type, exc_value, traceback)


class CFDatabaseOperations(SQLiteDatabaseOperations):
    pass

    # This patches some weird bugs related to the Database class
    def _quote_params_for_last_executed_query(self, params):
        """
        Only for last_executed_query! Don't use this to execute SQL queries!
        """
        # This function is limited both by SQLITE_LIMIT_VARIABLE_NUMBER (the
        # number of parameters, default = 999) and SQLITE_MAX_COLUMN (the
        # number of return values, default = 2000). Since Python's sqlite3
        # module doesn't expose the get_limit() C API, assume the default
        # limits are in effect and split the work in batches if needed.
        BATCH_SIZE = 999
        if len(params) > BATCH_SIZE:
            results = ()
            for index in range(0, len(params), BATCH_SIZE):
                chunk = params[index: index + BATCH_SIZE]
                results += self._quote_params_for_last_executed_query(chunk)
            return results

        sql = "SELECT " + ", ".join(["QUOTE(?)"] * len(params))
        # Bypass Django's wrappers and use the underlying sqlite3 connection
        # to avoid logging this query - it would trigger infinite recursion.

        cursor = self.connection.connection.cursor()
        # Native sqlite3 cursors cannot be used as context managers.
        # try:
        #     return cursor.execute(sql, params).fetchone()
        # finally:
        #     cursor.close()

    def last_executed_query(self, cursor, sql, params):
        # Python substitutes parameters in Modules/_sqlite/cursor.c with:
        # bind_parameters(state, self->statement, parameters);
        # Unfortunately there is no way to reach self->statement from Python,
        # so we quote and substitute parameters manually.
        if params:
            if isinstance(params, (list, tuple)):
                params = self._quote_params_for_last_executed_query(params)
            else:
                values = tuple(params.values())
                values = self._quote_params_for_last_executed_query(values)
                params = dict(zip(params, values))
            try:
                return sql % params
            except:
                return sql
        # For consistency with SQLiteCursorWrapper.execute(), just return sql
        # when there are no parameters. See #13648 and #17158.
        else:
            return sql

    def bulk_insert_sql(self, fields, placeholder_rows):
        placeholder_rows_sql = (", ".join(row) for row in placeholder_rows)
        values_sql = ", ".join("(%s)" % sql for sql in placeholder_rows_sql)
        return "VALUES " + values_sql


class CFDatabaseFeatures(SQLiteDatabaseFeatures):
    has_select_for_update = True
    has_native_uuid_field = False
    atomic_transactions = False
    supports_transactions = False
    can_release_savepoints = False
    supports_atomic_references_rename = False
    can_clone_databases = False
    can_rollback_ddl = False
    # Unsupported add column and foreign key in single statement
    # https://github.com/pingcap/tidb/issues/45474
    can_create_inline_fk = False
    order_by_nulls_first = True
    create_test_procedure_without_params_sql = None
    create_test_procedure_with_int_param_sql = None
    supports_aggregate_filter_clause = True
    can_defer_constraint_checks = False
    supports_pragma_foreign_key_check = False
    can_alter_table_rename_column = False
    max_query_params = 100
    can_clone_databases = False
    can_rollback_ddl = False
    supports_atomic_references_rename = False
    supports_forward_references = False
    supports_transactions = False
    has_bulk_insert = True
    # supports_select_union = False
    # supports_select_intersection = False
    # supports_select_difference = False
    can_return_columns_from_insert = True

    minimum_database_version = (4,)


class CFResult:
    lastrowid = None
    rowcount = -1

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        return iter(self.data)

    def set_lastrowid(self, value):
        self.lastrowid = value

    def set_rowcount(self, value):
        self.rowcount = value

    def fetchone(self):
        if len(self.data) > 0:
            return self.data.pop()
        return None

    def fetchall(self):
        ret = []
        while True:
            row = self.fetchone()
            if row is None:
                break
            ret.append(row)
        return ret

    def fetchmany(self, size=1):
        ret = []
        while size > 0:
            row = self.fetchone()
            if row is None:
                break
            ret.append(row)
            if size is not None:
                size -= 1

        return ret

    @staticmethod
    def from_object(query, params, data, rows_read=None, rows_written=None, last_row_id=None):
        try:
            from pyodide.ffi import jsnull
        except ImportError:
            jsnull = None

        result = []

        for row in data:
            row_items = ()
            if isinstance(row, list):
                for v in row:
                    if v is jsnull:
                        row_items += (None,)
                    else:
                        row_items += (v,)
            else:
                for k, v in row.items():
                    if v is jsnull:
                        row_items += (None,)
                    else:
                        row_items += (v,)

            result.append(row_items)

        instance = CFResult(result)

        if rows_read or rows_written:
            if "INSERT" in query.upper():
                instance.set_rowcount(rows_written or 0)
            elif "UPDATE" in query.upper() or "DELETE" in query.upper():
                instance.set_rowcount(rows_written or 0)
            else:
                instance.set_rowcount(rows_read or 0)

        if last_row_id is not None:
            instance.set_lastrowid(last_row_id)

        return instance


class CFDatabase:
    def __init__(self, database_wrapper):
        self.databaseWrapper = database_wrapper

    DataError = DataError

    OperationalError = OperationalError

    IntegrityError = IntegrityError

    InternalError = InternalError

    ProgrammingError = ProgrammingError

    NotSupportedError = NotSupportedError
    DatabaseError = DatabaseError
    InterfaceError = InterfaceError
    Error = Error

    _defer_foreign_keys = False

    lastResult: CFResult = None

    def defer_foreign_keys(self, state):
        _defer_foreign_keys = state

    @staticmethod
    def connect(binding):
        return CFDatabase(binding)

    def cursor(self):
        return self

    def commit(self):
        return  # No commits allowed

    def rollback(self):
        return  # No commits allowed

    def fetchone(self):
        return self.lastResult.fetchone()

    def fetchall(self):
        return self.lastResult.fetchall()

    def fetchmany(self, size=1):
        return self.lastResult.fetchmany(size)

    @property
    def lastrowid(self):
        return self.lastResult.lastrowid

    @property
    def rowcount(self):
        return self.lastResult.rowcount

    def execute(self, query, params=None) -> None:
        if params:
            newParams = []
            for v in list(params):
                if v is True:
                    v = 1
                elif v is False:
                    v = 0

                newParams.append(v)

            params = tuple(newParams)

        self.lastResult = self.databaseWrapper.run_query(query, params)

        return self

    def close(self):
        return


def is_read_only_query(query: str) -> bool:
    parsed = sqlparse.parse(query.strip())

    if not parsed:
        return False  # Invalid or empty query

    # Get the first statement
    statement = parsed[0]

    # Check if the statement is a SELECT query
    if statement.get_type().upper() == "SELECT":
        return True

    # List of modifying query types
    modifying_types = {"INSERT", "UPDATE", "DELETE", "CREATE", "ALTER", "DROP", "REPLACE"}

    return statement.get_type().upper() not in modifying_types


class CFSQLCompiler(SQLCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super().as_sql(with_limits=with_limits, with_col_aliases=with_col_aliases)
        return sql, params

    def compile_func(self, func, **extra_context):
        if isinstance(func, (TruncYear, TruncQuarter, TruncMonth, TruncWeek, TruncDay, TruncHour, TruncMinute, TruncSecond, TruncDate, TruncTime)):
            return self._compile_date_trunc(func, **extra_context)
        return super().compile_func(func, **extra_context)

    def _compile_date_trunc(self, func, **extra_context):
        kind = func.kind
        field_sql, params = self.compile(func.source_expressions[0])

        templates = {
            'year': 'STRFTIME("%Y-01-01", {})'.format(field_sql),
            'quarter': 'CASE CAST(STRFTIME("%%m", {}) AS INTEGER) WHEN 1 THEN STRFTIME("%%Y-01-01", {}) WHEN 2 THEN STRFTIME("%%Y-04-01", {}) WHEN 3 THEN STRFTIME("%%Y-07-01", {}) WHEN 4 THEN STRFTIME("%%Y-10-01", {}) END'.format(field_sql, field_sql, field_sql, field_sql, field_sql),
            'month': 'STRFTIME("%Y-%m-01", {})'.format(field_sql),
            'week': 'DATE({}, "-" || CAST((CAST(STRFTIME("%w", {}) AS INTEGER) + 6) %% 7 AS TEXT) || " days")'.format(field_sql, field_sql),
            'day': 'DATE({})'.format(field_sql),
            'hour': 'STRFTIME("%Y-%m-%d %H:00:00", {})'.format(field_sql),
            'minute': 'STRFTIME("%Y-%m-%d %H:%M:00", {})'.format(field_sql),
            'second': 'STRFTIME("%Y-%m-%d %H:%M:%S", {})'.format(field_sql),
            'date': 'DATE({})'.format(field_sql),
            'time': 'TIME({})'.format(field_sql),
        }

        sql = templates.get(kind, field_sql)
        return sql, params


class CFDatabaseWrapper(SQLiteDatabaseWrapper):
    # this is defined in the class extending this one
    # vendor = "cloudflare_d1"
    # display_name = "D1"

    Database = CFDatabase
    SchemaEditorClass = CFDatabaseSchemaEditor
    client_class = CFDatabaseClient
    creation_class = CFDatabaseCreation

    # Classes instantiated in __init__().
    features_class = CFDatabaseFeatures
    introspection_class = CFDatabaseIntrospection
    ops_class = CFDatabaseOperations

    transaction_modes = frozenset([])

    def get_compiler(self, default_using=None, using=None, **kwargs):
        if using is None:
            using = default_using
        return CFSQLCompiler(self.get_query(), self, using, **kwargs)

    def get_database_version(self):
        return (4,)

    def get_connection_params(self):
        raise NotImplementedError()

    def get_new_connection(self, conn_params):
        conn = CFDatabase.connect(self)
        return conn

    def create_cursor(self, name=None):
        return self.connection.cursor()

    def close(self):
        return

    def _savepoint_allowed(self):
        return False

    def _set_autocommit(self, commit):
        return

    def set_autocommit(
            self, autocommit, force_begin_transaction_with_broken_autocommit=False
    ):
        return

    def disable_constraint_checking(self):
        self.cursor().defer_foreign_keys(False)
        return True

    def enable_constraint_checking(self):
        self.cursor().defer_foreign_keys(True)

    def is_usable(self):
        return True

    def run_query(self, query, params=None) -> CFResult:
        raise NotImplementedError()



from .storage import get_storage
from ...base_engine import CFDatabaseWrapper, CFResult


class DatabaseWrapper(CFDatabaseWrapper):
    vendor = "cloudflare_durable_objects"
    display_name = "DO"
    binding: str

    def get_connection_params(self):
        return {}


    def process_query(self, query, params=None):
        if params is None:
            query = query.replace('%s', '?')
        else:
            new_params = []
            for param in params:
                if param is None:
                    query = query.replace('%s', 'null', 1)
                else:
                    new_params.append(param)
                    query = query.replace('%s', '?', 1)

            params = new_params


        if self.cursor()._defer_foreign_keys:
            return f'''
            PRAGMA defer_foreign_keys = on

            {query}

            PRAGMA defer_foreign_keys = off
            '''

        return query, params

    def run_query(self, query, params=None) -> CFResult:
        proc_query, params = self.process_query(query, params)

        db = get_storage()

        if params:
            stmt = db.exec(proc_query, *params);
        else:
            stmt = db.exec(proc_query);

        try:
            response = stmt.raw().toArray().to_py()
            result = CFResult.from_object(query, params, response, stmt.rowsRead, stmt.rowsWritten)
        except:
            from js import Error
            Error.stackTraceLimit = 1e10
            raise Error(Error.new().stack)

        return result



import os
from io import BytesIO


async def handle_wsgi(request, app):
    os.environ.setdefault('DJANGO_ALLOW_ASYNC_UNSAFE', 'false')
    from js import Object, Response, URL, console

    url = URL.new(request.url)
    assert url.protocol[-1] == ":"
    scheme = url.protocol[:-1]
    path = url.pathname
    assert "?".startswith(url.search[0:1])
    query_string = url.search[1:]
    method = str(request.method).upper()

    host = url.host.split(':')[0]

    wsgi_request = {
        'REQUEST_METHOD': method,
        'PATH_INFO': path,
        'QUERY_STRING': query_string,
        'SERVER_NAME': host,
        'SERVER_PORT': url.port,
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'wsgi.input': BytesIO(b''),
        'wsgi.errors': console.error,
        'wsgi.version': (1, 0),
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': True,
        'wsgi.url_scheme': scheme,
    }

    if request.headers.get('content-type'):
        wsgi_request['CONTENT_TYPE'] = request.headers.get('content-type')

    if request.headers.get('content-length'):
        wsgi_request['CONTENT_LENGTH'] = request.headers.get('content-length')

    for header in request.headers.items():
        wsgi_request[f'HTTP_{header[0].upper()}'] = header[1]

    if method in ['POST', 'PUT', 'PATCH']:
        body = (await request._js_request.arrayBuffer()).to_bytes()
        wsgi_request['wsgi.input'] = BytesIO(body)

    def start_response(status_str, response_headers):
        nonlocal status, headers
        status = status_str
        headers = response_headers

    resp = app(wsgi_request, start_response)
    status = resp.status_code
    headers = resp.headers

    final_response = Response.new(
        resp.content.decode('utf-8'), headers=Object.fromEntries(headers.items()), status=status
    )

    for k, v in resp.cookies.items():
        value = str(v)
        final_response.headers.set('Set-Cookie', value.replace('Set-Cookie: ', '', 1));

    return final_response


class DjangoCF:
    def get_app(self):
        raise NotImplementedError("Please implement implement get_app in your django_cf worker")

    async def fetch(self, request):
        return await handle_wsgi(request, self.get_app())


class DjangoCFDurableObject:
    def get_app(self):
        raise NotImplementedError("Please implement implement get_app in your django_cf worker")

    def __init__(self, ctx, env):
        self.ctx = ctx
        self.env = env

        from django_cf.db.backends.do.storage import set_storage
        set_storage(self.ctx.storage.sql)

    def fetch(self, request):
        return handle_wsgi(request, self.get_app())



import os
from datetime import datetime
from io import BytesIO
from django.core.files.storage import Storage
from django.core.files.base import File
from django.utils.deconstruct import deconstructible
from js import Uint8Array

class R2File(File):
    """
    A file-like object for R2 storage.
    """
    def __init__(self, name, storage, mode='rb'):
        self.name = name
        self._storage = storage
        self._mode = mode
        self._file = None

    @property
    def file(self):
        if self._file is None:
            content = self._storage._read(self.name)
            self._file = BytesIO(content) if content is not None else BytesIO()
        return self._file

    def read(self, num_bytes=None):
        return self.file.read(num_bytes)

    def write(self, content):
        if 'w' not in self._mode and 'a' not in self._mode:
            raise AttributeError("File not opened for writing")
        return self.file.write(content)

    def close(self):
        if self._file is not None:
            self._file.close()


@deconstructible
class R2Storage(Storage):
    """
    Django storage backend for Cloudflare R2.

    Configuration in Django settings:
        STORAGES = {
            "default": {
                "BACKEND": "django_cf.storage.R2Storage",
                "OPTIONS": {
                    "binding": "BUCKET",  # The R2 binding name from wrangler.jsonc
                    "location": "",  # Optional prefix for all files
                }
            }
        }

    In your wrangler.jsonc, configure the R2 bucket binding:
        {
            "r2_buckets": [
                {
                    "binding": "BUCKET",
                    "bucket_name": "<YOUR_BUCKET_NAME>"
                }
            ]
        }
    """

    def __init__(self, binding='BUCKET', location='', allow_overwrite=False):
        self.binding = binding
        self.location = location.strip('/')
        self.allow_overwrite = allow_overwrite
        self._bucket = None
        self._import_from_javascript = None
        self._run_sync = None

    def _get_bucket(self):
        """Lazy initialization of the R2 bucket binding."""
        if self._bucket is None:
            if self._import_from_javascript is None:
                try:
                    from workers import import_from_javascript
                    from pyodide.ffi import run_sync
                    self._import_from_javascript = import_from_javascript
                    self._run_sync = run_sync

                except ImportError as e:
                    raise Exception("Code not running inside a worker!")

            cf_workers = self._import_from_javascript("cloudflare:workers")
            self._bucket = getattr(cf_workers.env, self.binding)

        return self._bucket

    def _full_path(self, name):
        """Generate the full path including location prefix."""
        if self.location:
            return f"{self.location}/{name}"
        return name

    def _open(self, name, mode='rb'):
        """
        Retrieve the file from R2.
        """
        return R2File(name, self, mode)

    def _read(self, name):
        """
        Read the content of a file from R2.
        """
        full_path = self._full_path(name)
        try:
            bucket = self._get_bucket()
            r2_object = self._run_sync(bucket.get(full_path))

            if r2_object is None:
                return None

            return self._run_sync(r2_object.arrayBuffer()).to_bytes()
        except Exception:
            return None

    def _save(self, name, content):
        """
        Save a file to R2.
        """
        full_path = self._full_path(name)

        if hasattr(content, 'read'):
            file_content = content.read()
            file_content = Uint8Array.new(file_content)
        else:
            file_content = content

        bucket = self._get_bucket()

        options = {}
        if hasattr(content, 'content_type') and content.content_type:
            options['httpMetadata'] = {'contentType': content.content_type}

        self._run_sync(bucket.put(full_path, file_content, options if options else None))
        return name

    def delete(self, name):
        """
        Delete a file from R2.
        """
        full_path = self._full_path(name)
        bucket = self._get_bucket()
        self._run_sync(bucket.delete(full_path))

    def exists(self, name):
        """
        Check if a file exists in R2.
        """
        full_path = self._full_path(name)
        try:
            bucket = self._get_bucket()
            result = self._run_sync(bucket.head(full_path)).to_py()
            return result is not None
        except Exception:
            return False

    def listdir(self, path):
        """
        List the contents of a directory in R2.

        Returns:
            tuple: (directories, files)
        """
        full_path = self._full_path(path)
        if full_path and not full_path.endswith('/'):
            full_path += '/'

        bucket = self._get_bucket()
        result = self._run_sync(bucket.list({'prefix': full_path, 'delimiter': '/'})).to_py()

        directories = []
        files = []

        delimited_prefixes = result.get('delimitedPrefixes', [])
        for delimited_prefix in delimited_prefixes:
            directories.append(os.path.basename(delimited_prefix.replace(full_path, "", 1).rstrip('/')))


        objects = result.get('objects', [])
        for obj in objects:
            _obj = obj.to_py()

            if not obj.key.endswith('/'):
                files.append(os.path.basename(obj.key))

        return directories, files

    def size(self, name):
        """
        Return the size of a file in bytes.
        """
        full_path = self._full_path(name)
        try:
            bucket = self._get_bucket()
            metadata = self._run_sync(bucket.head(full_path)).to_py()
            if metadata and hasattr(metadata, 'size'):
                return metadata.size
            return 0
        except Exception:
            return 0

    def url(self, name):
        """
        Return the URL for accessing the file.
        
        Uses Django's MEDIA_URL setting to construct the file URL.
        Ensure your web server is configured to serve files from R2 at the MEDIA_URL path.
        """
        from django.conf import settings
        
        if not hasattr(settings, 'MEDIA_URL') or not settings.MEDIA_URL:
            raise ValueError(
                "MEDIA_URL must be configured in Django settings to use R2Storage. "
                "Configure your web server to proxy requests from MEDIA_URL to your R2 bucket."
            )
        
        full_path = self._full_path(name)
        media_url = settings.MEDIA_URL.rstrip('/')
        return f"{media_url}/{full_path}"

    def get_accessed_time(self, name):
        """
        Return the last accessed time of a file.
        R2 doesn't provide access time, so return modified time.
        """
        return self.get_modified_time(name)

    def get_created_time(self, name):
        """
        Return the creation time of a file.
        R2 doesn't provide creation time, so return modified time.
        """
        return self.get_modified_time(name)

    def get_modified_time(self, name):
        """
        Return the last modified time of a file.
        """
        full_path = self._full_path(name)
        try:
            bucket = self._get_bucket()
            metadata = self._run_sync(bucket.head(full_path)).to_py()

            if metadata and hasattr(metadata, 'uploaded'):
                uploaded = metadata.uploaded
                if isinstance(uploaded, datetime):
                    return uploaded
                return datetime.fromisoformat(str(uploaded))

            return datetime.now()
        except Exception:
            return datetime.now()

    def get_available_name(self, name, max_length=None):
        """
        Return a filename that's available in R2.
        """
        if max_length and len(name) > max_length:
            raise Exception(f"File name is too long (max {max_length} characters)")

        if self.allow_overwrite:
            return name

        if not self.exists(name):
            return name

        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        count = 1

        while self.exists(name) and (max_length is None or len(name) <= max_length):
            name = os.path.join(dir_name, f"{file_root}_{count}{file_ext}")
            count += 1

        return name



from django.core.exceptions import ImproperlyConfigured

from ...base_engine import CFDatabaseWrapper, is_read_only_query, CFResult


class DatabaseWrapper(CFDatabaseWrapper):
    vendor = "cloudflare_d1"
    display_name = "D1"
    binding: str

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if not settings_dict["CLOUDFLARE_BINDING"]:
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the CLOUDFLARE_BINDING value."
            )
        kwargs = {
            "binding": settings_dict["CLOUDFLARE_BINDING"],
        }
        return kwargs

    def get_new_connection(self, conn_params):
        self.binding = conn_params["binding"]
        return super().get_new_connection(conn_params)

    def __init__(self, *args):
        super().__init__(*args)

        try:
            from workers import import_from_javascript
            from pyodide.ffi import run_sync
            self.import_from_javascript = import_from_javascript
            self.run_sync = run_sync
        except ImportError as e:
            print(e)
            raise Exception("Code not running inside a worker!")


    def process_query(self, query, params=None):
        if params is None:
            query = query.replace('%s', '?')
        else:
            new_params = []
            for param in params:
                if param is None:
                    query = query.replace('%s', 'null', 1)
                else:
                    new_params.append(param)
                    query = query.replace('%s', '?', 1)

            params = new_params


        if self.cursor()._defer_foreign_keys:
            return f'''
            PRAGMA defer_foreign_keys = on

            {query}

            PRAGMA defer_foreign_keys = off
            '''

        return query, params

    def run_query(self, query, params=None) -> CFResult:
        proc_query, params = self.process_query(query, params)

        cf_workers = self.import_from_javascript("cloudflare:workers")
        db = getattr(cf_workers.env, self.binding)

        if params:
            stmt = db.prepare(proc_query).bind(*params);
        else:
            stmt = db.prepare(proc_query);

        read_only = is_read_only_query(proc_query)
        try:
            if read_only:
                response = self.run_sync(stmt.raw()).to_py()
                result = CFResult.from_object(query, params, response, len(response), 0)
            else:
                response = self.run_sync(stmt.all())
                result = CFResult.from_object(query, params, response.results.to_py(), response.meta.rows_read, response.meta.rows_written,
                                            response.meta.last_row_id)
        except:
            from js import Error
            Error.stackTraceLimit = 1e10
            raise Error(Error.new().stack)

        return result



import requests
from datetime import datetime, timedelta
from .utils import r2_web_server  # NOQA


def test_date_trunc_month(r2_web_server):
    """Test truncating dates to month."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 2'}, timeout=10)
    
    # Test month truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'month'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'month'
    assert len(result['results']) > 0
    assert all(isinstance(r, str) for r in result['results'])


def test_date_trunc_year(r2_web_server):
    """Test truncating dates to year."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    
    # Test year truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'year'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'year'
    assert len(result['results']) > 0


def test_date_trunc_day(r2_web_server):
    """Test truncating dates to day."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    
    # Test day truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'day'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'day'
    assert len(result['results']) > 0


def test_date_trunc_quarter(r2_web_server):
    """Test truncating dates to quarter."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    
    # Test quarter truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'quarter'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'quarter'
    assert len(result['results']) > 0


def test_date_trunc_week(r2_web_server):
    """Test truncating dates to week."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    
    # Test week truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'week'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'week'
    assert len(result['results']) > 0


def test_date_trunc_hour(r2_web_server):
    """Test truncating dates to hour."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    
    # Create test events
    requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Event 1'}, timeout=10)
    
    # Test hour truncation
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'hour'}, timeout=10)
    
    assert response.status_code == 200
    result = response.json()
    assert result['status'] == 'success'
    assert result['type'] == 'hour'
    assert len(result['results']) > 0


def test_date_trunc_invalid_type(r2_web_server):
    """Test that invalid truncation type returns error."""
    base_url = r2_web_server.base_url
    
    # Test invalid truncation type
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'invalid'}, timeout=10)
    
    assert response.status_code == 400
    result = response.json()
    assert result['status'] == 'error'
    assert 'Invalid truncation type' in result['message']


def test_date_trunc_create_and_clear_events(r2_web_server):
    """Test creating and clearing events."""
    base_url = r2_web_server.base_url
    
    # Clear any existing events
    response = requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    assert response.status_code == 200
    assert response.json()['status'] == 'success'
    
    # Create an event
    response = requests.post(f"{base_url}/__date_trunc_create_event__/", data={'name': 'Test Event'}, timeout=10)
    assert response.status_code == 200
    assert response.json()['status'] == 'success'
    
    # Verify event was created by querying
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'month'}, timeout=10)
    assert response.status_code == 200
    assert len(response.json()['results']) > 0
    
    # Clear events
    response = requests.post(f"{base_url}/__date_trunc_clear_events__/", timeout=10)
    assert response.status_code == 200
    
    # Verify events were cleared
    response = requests.get(f"{base_url}/__date_trunc_test__/", params={'type': 'month'}, timeout=10)
    assert response.status_code == 200
    assert len(response.json()['results']) == 0



import json
import base64
import hashlib
import time
import urllib.request
import urllib.error
from django.contrib.auth import get_user_model, login
from django.contrib.auth.models import AnonymousUser
from django.http import JsonResponse
from django.conf import settings
from django.core.cache import cache
import logging

logger = logging.getLogger(__name__)

User = get_user_model()

try:
    from js import fetch
    from pyodide.ffi import run_sync

    IS_WORKER = True
except ImportError:
    IS_WORKER = False
    fetch = None
    run_sync = None


class CloudflareAccessMiddleware:
    """
    Django middleware for Cloudflare Access authentication.

    This middleware:
    1. Extracts JWT from CF-Access-Jwt-Assertion header or cf_authorization cookie
    2. Validates the JWT against Cloudflare's public keys
    3. Creates or retrieves user based on JWT claims
    4. Logs the user in automatically

    Settings required (at least one):
    - CLOUDFLARE_ACCESS_AUD: Your Cloudflare Access Application Audience (AUD) tag
    - CLOUDFLARE_ACCESS_TEAM_NAME: Your Cloudflare team name (e.g., 'yourteam')

    If only AUD is provided, team name will be extracted from JWT claims.
    If only team name is provided, AUD will be validated from JWT claims.
    """

    def __init__(self, get_response):
        self.get_response = get_response

        # Validate required settings - at least one must be provided
        self.aud = getattr(settings, 'CLOUDFLARE_ACCESS_AUD', None)
        self.team_name = getattr(settings, 'CLOUDFLARE_ACCESS_TEAM_NAME', None)

        if not self.aud and not self.team_name:
            raise ValueError("Either CLOUDFLARE_ACCESS_AUD or CLOUDFLARE_ACCESS_TEAM_NAME setting is required")

        # If team_name is provided, use it to construct the certs URL
        if self.team_name:
            self.team_domain = f"{self.team_name}.cloudflareaccess.com"
            self.certs_url = f"https://{self.team_domain}/cdn-cgi/access/certs"
        else:
            # We'll determine the team domain from the JWT later
            self.team_domain = None
            self.certs_url = None

        # Optional settings
        self.exempt_paths = getattr(settings, 'CLOUDFLARE_ACCESS_EXEMPT_PATHS', [])
        self.cache_timeout = getattr(settings, 'CLOUDFLARE_ACCESS_CACHE_TIMEOUT', 3600)  # 1 hour

    def __call__(self, request):
        # Always try to authenticate with Cloudflare Access if JWT is present
        # This ensures users stay logged in even after Django logout
        try:
            user = self._authenticate_cloudflare_access(request)
            if user:
                # Set the user on the request (this overrides any logged-out state)
                request.user = user
                # Clear any logout-related session data
                if hasattr(request, 'session'):
                    # User was logged out but CF Access is still valid, re-authenticate
                    try:
                        login(request, user)
                        logger.info("Logged in user %s", user)
                    except Exception as e:
                        logger.warning(f"Failed to re-login user after logout: {str(e)}")
            else:
                # Check if path is exempt from authentication
                if not self._is_exempt_path(request.path):
                    # Authentication failed and path is not exempt, return 401
                    return JsonResponse(
                        {'error': 'Cloudflare Access authentication required'},
                        status=401
                    )
                # Path is exempt, continue with anonymous user

        except Exception as e:
            logger.error(f"Cloudflare Access authentication error: {repr(e)}")
            # Only return 500 for non-exempt paths
            if not self._is_exempt_path(request.path):
                return JsonResponse(
                    {'error': 'Authentication error'},
                    status=500
                )

        return self.get_response(request)

    def _is_exempt_path(self, path):
        """Check if the current path is exempt from authentication."""
        for exempt_path in self.exempt_paths:
            if path.startswith(exempt_path):
                return True
        return False

    def _authenticate_cloudflare_access(self, request):
        """
        Authenticate user using Cloudflare Access JWT.
        Returns User object if valid, None otherwise.
        """
        # Extract JWT token from header or cookie
        jwt_token = self._extract_jwt_token(request)
        if not jwt_token:
            return None

        # If we don't have team_name, try to extract it from JWT first
        if not self.team_name:
            team_name = self._extract_team_name_from_jwt(jwt_token)
            if not team_name:
                logger.error("Unable to determine team name from JWT")
                return None
            self.team_domain = f"{team_name}.cloudflareaccess.com"
            self.certs_url = f"https://{self.team_domain}/cdn-cgi/access/certs"

        # Get Cloudflare public keys
        public_keys = self._get_cloudflare_public_keys()
        if not public_keys:
            logger.error("Failed to retrieve Cloudflare public keys")
            return None

        email = None
        name = None

        # Validate and decode JWT
        try:
            # Try each public key until one works
            decoded_token = None
            for key_data in public_keys:
                try:
                    decoded_token = self._decode_and_verify_jwt(jwt_token, key_data)
                    if decoded_token:
                        break
                except Exception as e:
                    logger.debug(f"Key {key_data.get('kid')} failed: {str(e)}")
                    continue

            if not decoded_token:
                logger.warning("JWT token validation failed with all available keys")
                return None

            # Validate AUD if configured
            if self.aud:
                token_aud = decoded_token.get('aud')
                if isinstance(token_aud, list):
                    if self.aud not in token_aud:
                        logger.warning(f"JWT audience mismatch. Expected: {self.aud}, Got: {token_aud}")
                        return None
                elif token_aud != self.aud:
                    logger.warning(f"JWT audience mismatch. Expected: {self.aud}, Got: {token_aud}")
                    return None

            # Validate AUD if not configured but we have a team name
            if not self.aud and self.team_name:
                token_aud = decoded_token.get('aud')
                if not token_aud:
                    logger.warning("No audience found in JWT token")
                    return None

            # Extract user information from JWT claims
            email = decoded_token.get('email')
            name = decoded_token.get('name', '')

            # Try to get name from custom claims if not in standard claims
            if not name:
                custom_claims = decoded_token.get('custom', {})
                first_name = custom_claims.get('firstName', '')
                last_name = custom_claims.get('lastName', '')
                if first_name or last_name:
                    name = f"{first_name} {last_name}".strip()

            if not email:
                logger.warning("No email found in JWT token")
                return None

        except Exception as e:
            logger.warning(f"JWT token validation error: {repr(e)}")
            return None

        # Get or create user
        user = self._get_or_create_user(email, name)
        return user

    def _extract_jwt_token(self, request):
        """Extract JWT token from CF-Access-Jwt-Assertion header or cf_authorization cookie."""
        # Try header first (most common)
        jwt_token = request.META.get('HTTP_CF_ACCESS_JWT_ASSERTION')
        if jwt_token:
            return jwt_token

        # Try alternative header format
        jwt_token = request.META.get('HTTP_CF_ACCESS_JWT_ASSERTION'.replace('_', '-'))
        if jwt_token:
            return jwt_token

        # Try cookie
        jwt_token = request.COOKIES.get('CF_Authorization')
        if jwt_token:
            return jwt_token

        # Try alternative cookie name
        jwt_token = request.COOKIES.get('cf_authorization')
        if jwt_token:
            return jwt_token

        return None

    def _extract_team_name_from_jwt(self, jwt_token):
        """Extract team name from JWT token without validation (for bootstrapping)."""
        try:
            # Split JWT into parts
            parts = jwt_token.split('.')
            if len(parts) != 3:
                return None

            # Decode payload (add padding if needed)
            payload_part = parts[1]
            payload_part += '=' * (4 - len(payload_part) % 4)
            payload_bytes = base64.urlsafe_b64decode(payload_part)
            payload = json.loads(payload_bytes.decode('utf-8'))

            # Extract issuer (iss) claim - format: https://teamname.cloudflareaccess.com
            issuer = payload.get('iss')
            if not issuer:
                return None

            # Extract team name from issuer URL
            if issuer.startswith('https://') and issuer.endswith('.cloudflareaccess.com'):
                team_name = issuer.replace('https://', '').replace('.cloudflareaccess.com', '')
                return team_name

            return None
        except Exception as e:
            logger.error(f"Failed to extract team name from JWT: {str(e)}")
            return None

    def _get_cloudflare_public_keys(self):
        """Retrieve Cloudflare public keys, with caching."""
        # Use team_name for cache key, or extract from team_domain if available
        cache_key_team = self.team_name or (self.team_domain.split('.')[0] if self.team_domain else 'unknown')
        cache_key = f"cloudflare_access_keys_{cache_key_team}"
        cached_keys = cache.get(cache_key)

        if cached_keys:
            return cached_keys

        if IS_WORKER:
            response = run_sync(fetch(self.certs_url))
            if response.status == 200:
                data = run_sync(response.json()).to_py()
            else:
                logger.error(f"Failed to fetch Cloudflare keys: HTTP {response.status}")
                return None
        else:
            try:
                with urllib.request.urlopen(self.certs_url) as response:
                    if response.status == 200:
                        data = json.loads(response.read().decode('utf-8'))
                    else:
                        logger.error(f"Failed to fetch Cloudflare keys: HTTP {response.status}")
                        return None
            except urllib.error.URLError as e:
                logger.error(f"Network error fetching Cloudflare keys: {str(e)}")
                return None
            except Exception as e:
                logger.error(f"Unexpected error fetching Cloudflare keys: {str(e)}")
                return None

        keys = data.get('keys', [])

        # Process keys for JWT validation
        processed_keys = []
        for key_info in keys:
            if key_info.get('kty') == 'RSA':
                # Extract RSA components
                try:
                    processed_key = self._process_rsa_key(key_info)
                    if processed_key:
                        processed_keys.append(processed_key)
                except Exception as e:
                    logger.warning(f"Failed to process key {key_info.get('kid')}: {str(e)}")
                    continue

        # Cache the keys
        cache.set(cache_key, processed_keys, self.cache_timeout)
        return processed_keys

    def _process_rsa_key(self, key_info):
        """Process RSA key from JWK format to usable format."""
        try:
            # Extract RSA components from JWK
            n = key_info.get('n')  # modulus
            e = key_info.get('e')  # exponent
            kid = key_info.get('kid')

            if not n or not e:
                return None

            # Decode base64url encoded values
            n_bytes = self._base64url_decode(n)
            e_bytes = self._base64url_decode(e)

            return {
                'kid': kid,
                'n': int.from_bytes(n_bytes, 'big'),
                'e': int.from_bytes(e_bytes, 'big')
            }
        except Exception as e:
            logger.warning(f"Failed to process RSA key: {str(e)}")
            return None

    def _base64url_decode(self, data):
        """Decode base64url encoded data."""
        # Add padding if needed
        data += '=' * (4 - len(data) % 4)
        return base64.urlsafe_b64decode(data)

    def _decode_and_verify_jwt(self, jwt_token, key_data):
        """Decode and verify JWT token using RSA key."""
        try:
            # Split JWT into parts
            parts = jwt_token.split('.')
            if len(parts) != 3:
                raise ValueError("Invalid JWT format")

            header_part, payload_part, signature_part = parts

            # Decode header
            header_bytes = self._base64url_decode(header_part)
            header = json.loads(header_bytes.decode('utf-8'))

            # Check if this key matches the kid in the header
            if header.get('kid') != key_data.get('kid'):
                raise ValueError("Key ID mismatch")

            # Check algorithm
            if header.get('alg') != 'RS256':
                raise ValueError("Unsupported algorithm")

            # Decode payload
            payload_bytes = self._base64url_decode(payload_part)
            payload = json.loads(payload_bytes.decode('utf-8'))

            # Check expiration
            exp = payload.get('exp')
            if exp and exp < time.time():
                raise ValueError("Token expired")

            # Check not before
            nbf = payload.get('nbf')
            if nbf and nbf > time.time():
                raise ValueError("Token not yet valid")

            # Verify signature
            message = f"{header_part}.{payload_part}".encode('utf-8')
            signature = self._base64url_decode(signature_part)

            if not self._verify_rsa_signature(message, signature, key_data):
                raise ValueError("Invalid signature")

            return payload

        except Exception as e:
            logger.debug(f"JWT verification failed: {str(e)}")
            return None

    def _verify_rsa_signature(self, message, signature, key_data):
        """Verify RSA signature using PKCS#1 v1.5 with SHA-256."""
        try:
            # Hash the message
            hash_obj = hashlib.sha256()
            hash_obj.update(message)
            message_hash = hash_obj.digest()

            # Create PKCS#1 v1.5 padding for SHA-256
            # DigestInfo for SHA-256: 30 31 30 0d 06 09 60 86 48 01 65 03 04 02 01 05 00 04 20
            digest_info = bytes.fromhex('3031300d060960864801650304020105000420')
            padded_hash = digest_info + message_hash

            # RSA signature verification
            n = key_data['n']
            e = key_data['e']

            # Convert signature to integer
            sig_int = int.from_bytes(signature, 'big')

            # RSA verification: sig^e mod n
            decrypted_int = pow(sig_int, e, n)

            # Convert back to bytes with proper padding
            # The key length in bytes
            key_length = (n.bit_length() + 7) // 8

            # Ensure we have the right number of bytes
            decrypted_bytes = decrypted_int.to_bytes(key_length, 'big')

            # Check minimum length
            if len(decrypted_bytes) < len(padded_hash) + 11:
                logger.debug(f"Decrypted bytes too short: {len(decrypted_bytes)} < {len(padded_hash) + 11}")
                return False

            # PKCS#1 v1.5 padding format: 0x00 0x01 [0xFF padding] 0x00 [DigestInfo + Hash]
            if len(decrypted_bytes) == 0 or decrypted_bytes[0] != 0x00:
                logger.debug(f"Invalid padding: first byte is {decrypted_bytes[0]:02x}, expected 0x00")
                return False

            if len(decrypted_bytes) < 2 or decrypted_bytes[1] != 0x01:
                logger.debug(f"Invalid padding: second byte is {decrypted_bytes[1]:02x}, expected 0x01")
                return False

            # Find the 0x00 separator
            separator_idx = -1
            for i in range(2, len(decrypted_bytes)):
                if decrypted_bytes[i] == 0x00:
                    separator_idx = i
                    break
                elif decrypted_bytes[i] != 0xFF:
                    logger.debug(f"Invalid padding byte at position {i}: {decrypted_bytes[i]:02x}, expected 0xFF")
                    return False

            if separator_idx == -1:
                logger.debug("No separator found in padding")
                return False

            # Ensure minimum padding length (at least 8 bytes of 0xFF)
            if separator_idx - 2 < 8:
                logger.debug(f"Padding too short: {separator_idx - 2} < 8")
                return False

            # Extract and compare the hash
            extracted_hash = decrypted_bytes[separator_idx + 1:]

            if len(extracted_hash) != len(padded_hash):
                logger.debug(f"Hash length mismatch: {len(extracted_hash)} != {len(padded_hash)}")
                return False

            result = extracted_hash == padded_hash
            if not result:
                logger.debug("Hash comparison failed")
                logger.debug(f"Expected: {padded_hash.hex()}")
                logger.debug(f"Got:      {extracted_hash.hex()}")

            return result

        except Exception as e:
            logger.debug(f"RSA signature verification failed: {str(e)}")
            return False

    def _get_or_create_user(self, email, name):
        """Get or create Django user from Cloudflare Access claims."""
        try:
            # Try to get existing user by email
            user = User.objects.get(email=email)

            # Update name if it has changed
            if name and user.get_full_name() != name:
                # Split name into first and last name
                name_parts = name.split(' ', 1)
                user.first_name = name_parts[0]
                user.last_name = name_parts[1] if len(name_parts) > 1 else ''
                user.save()

            return user

        except User.DoesNotExist:
            # Create new user
            name_parts = name.split(' ', 1) if name else ['', '']

            user = User.objects.create_user(
                username=email,  # Use email as username
                email=email,
                first_name=name_parts[0],
                last_name=name_parts[1] if len(name_parts) > 1 else '',
                is_active=True
            )

            logger.info(f"Created new user from Cloudflare Access: {email}")
            return user
